(function(global, factory) {
  typeof exports === "object" && typeof module !== "undefined" ? factory(exports, require("vue"), require("@resourge/shallow-clone")) : typeof define === "function" && define.amd ? define(["exports", "vue", "@resourge/shallow-clone"], factory) : (global = typeof globalThis !== "undefined" ? globalThis : global || self, factory(global.index = {}, global.vue, global.shallowClone));
})(this, function(exports2, vue, shallowClone) {
  "use strict";
  function isNumeric(value) {
    if (typeof value === "number") {
      return true;
    }
    return /^[-]?([1-9]\d*|0)(\.\d+)?$/.test(value);
  }
  const getKeyFromPaths = (paths) => {
    return paths.map((key) => `${isNumeric(key) ? `[${key}]` : `${key}`}`).join(".").replace(/\.\[/g, "[");
  };
  const formatErrors = (errors, defaultErrors = {}) => {
    const _simpleErrors = errors.reduce((obj, value) => {
      const keys = Array.isArray(value.path) ? value.path : value.path.split(".").map(
        (key) => key.split("[").map((arrayKey) => arrayKey.replace("]", ""))
      ).flat();
      const path = getKeyFromPaths(keys);
      if (!obj[path]) {
        obj[path] = [];
      }
      obj[path].push(...value.error !== void 0 ? [value.error] : value.errors);
      return obj;
    }, {});
    return {
      ...defaultErrors,
      ..._simpleErrors
    };
  };
  const createFormErrors = (onError) => (errors, defaultErrors = {}) => {
    return formatErrors(onError(errors), defaultErrors);
  };
  function isObject(value) {
    return toString.call(value) === "[object Object]";
  }
  const setValue = (field) => {
    return new Function("obj", "val", `obj${field ? `.${field}` : ""} = val`);
  };
  const getValue = (field) => {
    return new Function("obj", `return obj${field ? `.${field}` : ""}`);
  };
  const createGetterSetter = (field) => {
    return {
      set: setValue(field),
      get: getValue(field)
    };
  };
  const maxLimitOfGetterSetter = 1e4;
  const useGetterSetter = () => {
    const getterSetter = /* @__PURE__ */ new Map();
    const checkGetterSetter = (key) => {
      if (!getterSetter.has(key)) {
        if (getterSetter.size > maxLimitOfGetterSetter) {
          const firstElementKey = getterSetter.keys().next().value;
          if (firstElementKey) {
            getterSetter.delete(firstElementKey);
          }
        }
        getterSetter.set(key, createGetterSetter(key));
      }
    };
    const set = (key, form, value) => {
      checkGetterSetter(key);
      if (isObject(value) || Array.isArray(value)) {
        getterSetter.get(key).set(form, shallowClone.shallowClone(value));
      } else {
        getterSetter.get(key).set(form, value);
      }
    };
    const get = (key, form) => {
      checkGetterSetter(key);
      return getterSetter.get(key).get(form);
    };
    return {
      set,
      get
    };
  };
  const checkIfCanCheckError = (key, touches, onlyOnTouch, onlyOnTouchKeys = []) => {
    return !onlyOnTouch || (onlyOnTouch || onlyOnTouchKeys.length) && // @ts-expect-error no types
    (Boolean(touches[key]) || // @ts-expect-error no types
    onlyOnTouchKeys.some((onlyOnTouchKey) => Boolean(touches[onlyOnTouchKey])));
  };
  const useErrors = (state, formOptions) => {
    const cacheErrors = /* @__PURE__ */ new Map();
    const clearCacheErrors = () => {
      cacheErrors.clear();
    };
    const hasError = (key, options) => {
      const strict = (options == null ? void 0 : options.strict) ?? true;
      const onlyOnTouch = (options == null ? void 0 : options.onlyOnTouch) ?? (formOptions == null ? void 0 : formOptions.onlyOnTouchDefault) ?? true;
      const onlyOnTouchKeys = (options == null ? void 0 : options.onlyOnTouchKeys) ?? [];
      const newErrors = vue.computed(() => state.errors);
      const _errors = newErrors.value ?? {};
      let hasError2 = false;
      if (checkIfCanCheckError(key, state.touches, onlyOnTouch, onlyOnTouchKeys)) {
        hasError2 = Boolean(_errors[key]);
      }
      if (hasError2) {
        return true;
      } else {
        if (!strict) {
          const regex = new RegExp(`^${key.replace("[", "\\[").replace("]", "\\]")}`, "g");
          return Object.keys(_errors).some((errorKey) => {
            if (checkIfCanCheckError(errorKey, state.touches, onlyOnTouch)) {
              return regex.test(errorKey);
            }
            return false;
          });
        }
      }
      return hasError2;
    };
    function getErrors(key, options = {
      strict: true,
      onlyOnTouch: (formOptions == null ? void 0 : formOptions.onlyOnTouchDefault) ?? true,
      includeKeyInChildErrors: true,
      includeChildsIntoArray: false,
      onlyOnTouchKeys: []
    }) {
      const _strict = (options == null ? void 0 : options.strict) ?? true;
      const onlyOnTouch = (options == null ? void 0 : options.onlyOnTouch) ?? (formOptions == null ? void 0 : formOptions.onlyOnTouchDefault) ?? true;
      const onlyOnTouchKeys = (options == null ? void 0 : options.onlyOnTouchKeys) ?? [];
      const includeKeyInChildErrors = (options == null ? void 0 : options.includeKeyInChildErrors) ?? true;
      const includeChildsIntoArray = (options == null ? void 0 : options.includeChildsIntoArray) ?? false;
      const strict = includeChildsIntoArray ? false : _strict;
      const _errors = vue.computed(() => state.errors).value ?? {};
      const getErrors2 = (key2) => {
        if (checkIfCanCheckError(key2, state.touches, onlyOnTouch, onlyOnTouchKeys)) {
          return [..._errors[key2] ?? []];
        }
        return [];
      };
      const newErrors = getErrors2(key);
      if (!strict) {
        const regex = new RegExp(`^${key.replace("[", "\\[").replace("]", "\\]")}`, "g");
        Object.keys(_errors).forEach((errorKey) => {
          if (errorKey.includes(key) && errorKey !== key) {
            let newErrorKey = includeKeyInChildErrors === true ? errorKey : errorKey.replace(regex, "") || key;
            if (newErrorKey[0] === ".") {
              newErrorKey = newErrorKey.substring(1, newErrorKey.length);
            }
            const _newErrors = getErrors2(errorKey);
            if (includeChildsIntoArray) {
              newErrors.push(..._newErrors);
            }
            newErrors[newErrorKey] = _newErrors;
          }
        });
      }
      return newErrors;
    }
    return {
      hasError,
      getErrors,
      clearCacheErrors
    };
  };
  const useForm = (defaultValues, options = {}) => {
    const { validate, validateDefault = false } = options;
    const state = vue.reactive({
      errors: {},
      form: defaultValues,
      touches: {}
    });
    const isValid = vue.ref(false);
    const getterSetter = useGetterSetter();
    const { clearCacheErrors, getErrors, hasError } = useErrors(state, options);
    const onErrors = createFormErrors((errors2) => errors2);
    const validateForm = async (forceValidation) => {
      if (validate) {
        const changedKeys = Object.keys(state.touches);
        const result = await validate(state.form, changedKeys);
        const newErrors = onErrors(result);
        state.errors = newErrors;
        if (forceValidation) {
          Object.keys(newErrors).forEach((key) => {
            state.touches[key] = true;
          });
        }
        isValid.value = Object.keys(newErrors).length === 0;
      }
    };
    const reset = async (newFrom) => {
      const keys = Object.keys(newFrom);
      keys.forEach((key) => {
        const newValue = newFrom[key];
        getterSetter.set(key, state.form, newValue);
      });
      isValid.value = false;
      state.touches = {};
    };
    if (validateDefault) {
      vue.watchEffect(() => {
        clearCacheErrors();
        validateForm(false);
      });
    }
    const handleSubmit = (onSubmit) => async (e) => {
      if (e) {
        e.preventDefault();
      }
      validateForm(true);
      if (isValid.value) {
        await onSubmit(state.form);
        resetIsDirty();
      }
    };
    const field = (key) => {
      const value = vue.computed({
        // @ts-expect-error no types
        get: () => getterSetter.get(key, state.form),
        set: (newValue) => {
          getterSetter.set(key, state.form, newValue);
          state.touches[key] = true;
          validateForm(false);
        }
      });
      const isDirty2 = vue.computed(() => {
        const hasFoundParentKey = Object.keys(state.touches).some((touchKey) => touchKey.includes(key));
        return hasFoundParentKey ?? state.touches[key];
      });
      const onChange = (newValue) => {
        value.value = newValue;
      };
      const getError = () => getErrors(key);
      return { value: value.value, onChange, getError, hasError: (options2) => hasError(key, options2), isDirty: isDirty2 };
    };
    const getValue2 = (key) => {
      return getterSetter.get(key, state.form);
    };
    const getIsDirtyField = (key) => {
      return field(key).isDirty.value;
    };
    const watch = (callback) => {
      vue.watch(state.touches, function() {
        const changedKeys = Object.keys(state.touches);
        callback(changedKeys);
      });
    };
    const errors = vue.computed(() => state.errors);
    const form = state.form;
    const isDirty = vue.computed(() => {
      const changedKeys = Object.keys(state.touches);
      return changedKeys.length > 0;
    });
    const resetIsDirty = () => {
      state.touches = {};
    };
    return {
      errors,
      field,
      form,
      isDirty,
      getErrors,
      handleSubmit,
      hasError,
      getValue: getValue2,
      getIsDirtyField,
      isValid,
      reset,
      watch,
      resetIsDirty
    };
  };
  exports2.useForm = useForm;
  Object.defineProperty(exports2, Symbol.toStringTag, { value: "Module" });
});
